import inspect
import ast
from typing import get_type_hints


def from_parent_caller():
    # +------------------------------------------+
    # |                  STACK                   |
    # +-----+------------------------------------|
    # | idx |  function                          |
    # +-----+------------------------------------+
    # |  0  |  this very function                |
    # |  1  |  autogui (the streamlit component) |
    # |  2  |  function defined by user          |
    # +-----+------------------------------------+
    caller_frame = inspect.stack()[2]

    caller_function = caller_frame.function

    caller_module = inspect.getmodule(caller_frame[0])
    caller_func_obj = getattr(caller_module, caller_function)

    docstring = inspect.getsource(caller_func_obj).split('"""')
    if isinstance(docstring,list) and len(docstring) > 1 and 'def ' in docstring[0]:
        docstring = docstring[1]
    else:
        docstring = None

    args = caller_frame.frame.f_locals

    name, invars, outvars = from_func(caller_func_obj)
    return name, invars, outvars, docstring, args


def from_func(f):
    sig = inspect.signature(f)

    name = f.__name__
    args = {p.name:p.annotation for p in sig.parameters.values()}

    inferred_return_vars = []
    for node in ast.walk(ast.parse(inspect.getsource(f))):
        if isinstance(node, ast.Return):
            if isinstance(node.value, ast.Name):
                inferred_return_vars.append(node.value.id)
            elif isinstance(node.value, ast.Tuple):
                inferred_return_vars = inferred_return_vars + [v.id for v in node.value.elts]

    return_vars_types = sig.return_annotation
    return_vars_types = return_vars_types if isinstance(return_vars_types,tuple) else (return_vars_types,)
    out_vars = dict(zip(inferred_return_vars, return_vars_types))
    return name, args, out_vars

def from_vars(**kwargs):
    schema = dict()

    for var in kwargs:
        schema[var] = type(kwargs[var])

    return schema

def from_io(io):
    if isinstance(io, tuple):
        return (from_vars(io[0]), from_vars(io[1]))
    
    return (
        {v:io[v] for v in io if v.split("out_")[0]!=""},
        {v:io[v] for v in io if v.split("out_")[0]==""}
    )


def readable(schema):
    if isinstance(schema, tuple):
        return (readable(schema[0]), readable(schema[1]))

    txt = [
        f"{var}:{schema[var]}"
        for var in schema
    ]
    txt = ",".join(txt)
    return txt


FLOAT_IN_OUT = (from_vars(number=0.0), from_vars(out_number=0.0))
FLOAT_IN_BOOL_OUT = (from_vars(number=0.0), from_vars(out_number=False))
INT_IN_OUT = (from_vars(number=0), from_vars(out_number=0))
INT_IN_BOOL_OUT = (from_vars(number=0), from_vars(out_number=False))




from pathlib import Path

import pytest
from e2e_utils import StreamlitRunner
from playwright.sync_api import Page, expect

ROOT_DIRECTORY = Path(__file__).parent.parent.absolute()
BASIC_EXAMPLE_FILE = ROOT_DIRECTORY / "my_component" / "example.py"


@pytest.fixture(autouse=True, scope="module")
def streamlit_app():
    with StreamlitRunner(BASIC_EXAMPLE_FILE) as runner:
        yield runner


@pytest.fixture(autouse=True, scope="function")
def go_to_app(page: Page, streamlit_app: StreamlitRunner):
    page.goto(streamlit_app.server_url)
    # Wait for app to load
    page.get_by_role("img", name="Running...").is_hidden()


def test_should_render_template(page: Page):
    frame_0 = page.frame_locator('iframe[title="my_component\\.my_component"]').nth(0)
    frame_1 = page.frame_locator('iframe[title="my_component\\.my_component"]').nth(1)

    st_markdown_0 = page.get_by_role("paragraph").nth(0)
    st_markdown_1 = page.get_by_role("paragraph").nth(1)

    expect(st_markdown_0).to_contain_text("You've clicked 0 times!")

    frame_0.get_by_role("button", name="Click me!").click()

    expect(st_markdown_0).to_contain_text("You've clicked 1 times!")
    expect(st_markdown_1).to_contain_text("You've clicked 0 times!")

    frame_1.get_by_role("button", name="Click me!").click()
    frame_1.get_by_role("button", name="Click me!").click()

    expect(st_markdown_0).to_contain_text("You've clicked 1 times!")
    expect(st_markdown_1).to_contain_text("You've clicked 2 times!")

    page.get_by_label("Enter a name").click()
    page.get_by_label("Enter a name").fill("World")
    page.get_by_label("Enter a name").press("Enter")

    expect(frame_1.get_by_text("Hello, World!")).to_be_visible()

    frame_1.get_by_role("button", name="Click me!").click()

    expect(st_markdown_0).to_contain_text("You've clicked 1 times!")
    expect(st_markdown_1).to_contain_text("You've clicked 3 times!")


def test_should_change_iframe_height(page: Page):
    frame = page.frame_locator('iframe[title="my_component\\.my_component"]').nth(1)

    expect(frame.get_by_text("Hello, Streamlit!")).to_be_visible()

    locator = page.locator('iframe[title="my_component\\.my_component"]').nth(1)

    page.wait_for_timeout(1000)
    init_frame_height = locator.bounding_box()["height"]
    assert init_frame_height != 0

    page.get_by_label("Enter a name").click()

    page.get_by_label("Enter a name").fill(35 * "Streamlit ")
    page.get_by_label("Enter a name").press("Enter")

    expect(frame.get_by_text("Streamlit Streamlit Streamlit")).to_be_visible()

    page.wait_for_timeout(1000)
    frame_height = locator.bounding_box()["height"]
    assert frame_height > init_frame_height

    page.set_viewport_size({"width": 150, "height": 150})

    expect(frame.get_by_text("Streamlit Streamlit Streamlit")).not_to_be_in_viewport()

    page.wait_for_timeout(1000)
    frame_height_after_viewport_change = locator.bounding_box()["height"]
    assert frame_height_after_viewport_change > frame_height



import streamlit as st
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error

# Load dataset
@st.cache
def load_data():
    return pd.read_csv('https://raw.githubusercontent.com/datasets/finance-vix/main/data/vix-daily.csv')

data = load_data()

# Dashboard title
st.title("Finance Analytics Dashboard")

# Display dataset
st.subheader("Dataset Preview")
st.write(data.head())

# Sidebar for user input
st.sidebar.header("Filter Options")
start_date = st.sidebar.date_input("Start Date", value=pd.to_datetime(data['Date'].min()))
end_date = st.sidebar.date_input("End Date", value=pd.to_datetime(data['Date'].max()))

# Filter data
filtered_data = data[(pd.to_datetime(data['Date']) >= start_date) & (pd.to_datetime(data['Date']) <= end_date)]

# Display filtered data
st.subheader("Filtered Dataset")
st.write(filtered_data)

# Visualization
st.subheader("Visualizations")

# Line Chart
st.write("VIX Closing Price Over Time")
fig, ax = plt.subplots()
ax.plot(pd.to_datetime(filtered_data['Date']), filtered_data['VIX Close'], label='VIX Close', color='blue')
ax.set_xlabel("Date")
ax.set_ylabel("VIX Close")
ax.legend()
plt.xticks(rotation=45)
st.pyplot(fig)

# Machine Learning Prediction
st.subheader("Machine Learning Prediction")

# Prepare data
filtered_data['Date'] = pd.to_datetime(filtered_data['Date'])
filtered_data['Date_ordinal'] = filtered_data['Date'].map(lambda x: x.toordinal())
X = filtered_data[['Date_ordinal']]
y = filtered_data['VIX Close']

# Split data
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)

# Train model
model = RandomForestRegressor(n_estimators=100, random_state=42)
model.fit(X_train, y_train)

# Predict
y_pred = model.predict(X_test)

# Display results
st.write("Mean Squared Error:", mean_squared_error(y_test, y_pred))

# Prediction Visualization
st.write("Predicted vs Actual Values")
fig, ax = plt.subplots()
ax.scatter(y_test, y_pred, color='green', label='Predicted vs Actual')
ax.plot([y_test.min(), y_test.max()], [y_test.min(), y_test.max()], color='red', label='Ideal Fit')
ax.set_xlabel("Actual Values")
ax.set_ylabel("Predicted Values")
ax.legend()
st.pyplot(fig)



import os
import json

from importlib.metadata import packages_distributions

import aisuite as ai

import importlib

client = ai.Client()

try:
    azure_client = getattr(importlib.import_module("openai"),"AzureOpenAI")(
        api_key=os.getenv("AZURE_API_KEY"),
        api_version=os.getenv("AZURE_API_VERSION"),
        azure_endpoint = os.getenv("AZURE_API_BASE")
    )
except:
    azure_client = None

def get_client(provider):
    return azure_client if provider.lower() == 'azure' else client


def completions(provider, model, messages):
    client = get_client(provider)

    # Azure is the exception, as it is not currently working with aisuite
    model=model if provider == 'azure' else f"{provider}:{model}"

    return client.chat.completions.create(model=model, messages=messages)

def get_available():
    # returns a dict:
    # - key: the name of the provider package that must be installed
    # - value: availability of that particular provider:
    #   - installed: whether package of that provider is installed
    #   - missing: env vars still to be set
    #   - env: all mandatory envs that must be set
    #   - default: model to be used by default, if any


    # requirements to validate:
    # - key: provider python packages to install
    # - value: mandatory env vars and default model if any

    module_dir = os.path.dirname(__file__)
    with open(os.path.join(module_dir, "providers.json"), "r") as f:
        providers = json.load(f)

    pkgs = packages_distributions()
    ready = {}
    not_ready = {}
    for provider in providers:
        p = providers[provider]

        if provider=='azure': # special case
            p['installed'] = azure_client != None
        else:
            p["installed"] = provider in pkgs

        p["missing"] = [v for v in p["env"] if v not in os.environ]

        warnings = []
        triage = ready

        if not p["installed"]:
            triage=not_ready
            warnings.append("Module not installed")

        if len(p["missing"]) > 0:
            triage=not_ready
            for env in p["missing"]:
                warnings.append(f"`{env}` is not set")
        
        p["warnings"] = warnings

        triage[provider] = p
                
    #return providers
    return dict(**ready, **not_ready)


def detect_default(s=None):
    providers = s.providers if s else get_available()
    for p in providers:
        if len(providers[p]['warnings']) == 0 and "default" in providers[p]:
            return p, providers[p]['default']

    return None, None



import re

from . import aiclient

s = None
GKEY = '_global'

class ddict(dict):
    def __getattr__(self, name):
        return self[name]
    
    def __setattr__(self, name, value):
        if isinstance(value, dict) and not isinstance(value, ddict):
            value = ddict(value)
        self[name] = value
    
    def __delattr__(self, name):
        del self[name]


def init(name, key, provider, model, patience, rerun, history, store_in, settings_key='autogui-settings'):
    actual_init = False
    if settings_key not in store_in:
        global s
        s = ddict({})
        store_in[settings_key] = s


    if "prefs" not in s:
        s.prefs = ddict({
            GKEY:ddict(instance="All")
            # keys related to local instances will fall here
        })

    key = key if key else re.sub("[^a-z0-9]","-",name.lower())
    if key not in s.prefs:
        actual_init = True
        s.prefs[key] = ddict(instance=name)


    if "providers" not in s:
        s.providers = aiclient.get_available()

    if actual_init: #not isset('provider',key):
        set_provider_model(provider, model, key)
        setpref("patience", patience, key)
        setpref("rerun", rerun, key)
        setpref("history", history, key) 

    return key, actual_init


def get_instance(key):
    return s.prefs[key]

def isset(setting, key=GKEY):
    return setting in s.prefs[key]

def _get(setting, key=GKEY):
    return s.prefs[key].get(setting, None)

def _set(setting, value, key=GKEY):
    s.prefs[key][setting] = value

def pop(setting, key=GKEY):
    if isset(setting, key=key):
        value = _get(setting, key=key)
        del s.prefs[key][setting]
        return value
    return None


def scope(setting,key=GKEY):
    # useful to track multiple settings that are dependent on one another
    return key if isset(setting,key) else GKEY

def getpref(setting=None, key=GKEY):
    if not setting:
        return s.prefs.get(key,None)
    return _get(setting, scope(setting,key))

def setpref(setting, value, key=GKEY, force_local=False):
    if value == None:
        if key!=GKEY:
            pop(setting, key)
        return
    
    key = GKEY if not isset(setting) else key
    if key==GKEY and not force_local:
        for key in s.prefs:
            if key==GKEY:
                _set(setting,value)
            elif isset(setting,key):
                #removes local setting, falling back to global
                pop(setting,key)
    elif force_local or value != _get(setting):
        _set(setting,value,key)
    else:
        pop(setting,key)


def get_provider_model(key=GKEY):
    provider = getpref('provider',key)
    sc = scope('provider', key)

    model = getpref('model',sc)

    return provider, model


def set_provider_model(provider, model, key=GKEY):
    if provider and model:
        key = GKEY if not isset('provider') else key
        if key==GKEY:
            for key in s.prefs:
                if key==GKEY:
                    _set('provider',provider)
                    _set('model',model)
                elif isset('provider',key):
                    #removes local setting, falling back to global
                    pop('provider',key)
                    pop('model',key)
        else:
            if provider == _get('provider') and model == _get('model'):
                #removes local setting, falling back to global
                pop('provider',key)
                pop('model',key)
            else:
                _set('provider',provider,key)
                _set('model',model,key)
    else:
        provider, model = aiclient.detect_default(s)
        set_provider_model(provider, model, key)





from openai import AzureOpenAI
import os
from . import schema, aiclient
import re
import hashlib

from importlib.metadata import packages_distributions

FUNCTION_NAME = "fcn"
FILE_NAME = "fcn.py"
SCHEMA = schema.FLOAT_IN_OUT

class InsufficientCapabilityError(Exception):
    def __init__(self, packages):
        message=f"Insufficient capabilities. Please install additional packages to enable this feature. Recommended are: {packages}."
        super().__init__(message)

IO = f"""

You are a coding assistant specialist in streamlit dashboards and have general
problem solving skills in python. Given a list of steps or features, your job
is to identify each step and then generate code to accomplish such tasks
sequentially. Provide UI elements for all parameters and/or inputs involved in
every step.

The code you generate will be included in function `{{FUNCTION_NAME}}`, which
takes `{{INPUT_SCHEMA}}` as input arguments and returns `{{OUTPUT_SCHEMA}}` as
output. Make sure to define unique keys for every streamlit component, but
never use any random function for it. Never use streamlit sidebar. Never use
any streamlit experimental feature. Never use caching.

Provide only code and nothing else. Never include markdown backticks. Prefer to
make use of the following packages: {{AVAILABLE_PKGS}}.

Never use streamlit titles or headers. If there are multiple steps, organize
those in expanders, tabs, or small subheaders. If there are few, or a single
task, do not enclose in any container, but simply render the due GUI elements.

Make sure to preserve the definition, input, and output of function
`{{FUNCTION_NAME}}`.

 """.replace("\n"," ")

VISUALIZATION = f"""

You are also a specialist in visualization. Make sure to use plots either when
explicitly requested, or only when necessary, to visualize some parameter
change, as a preview feature.

If images are associated to GUI components, make sure to organize in columns
(each group of preview and GUI in a set of columns), so any generated GUI is
relatable to the visualized result. Visualization on one side, GUI on the
other.

""".replace("\n"," ")

SYSTEM = f"{IO} {VISUALIZATION}"

def get_code(prompt, provider, model, system=SYSTEM, hist=None, compact_hist=False):
    # Templates first, containing variables
    system = system.format(
        IO=IO,
        VISUALIZATION=VISUALIZATION
    )

    # Then variables, which are possibly mentioned in templates
    system = system.format(
        FUNCTION_NAME=FUNCTION_NAME,
        INPUT_SCHEMA=schema.readable(SCHEMA[0]),
        OUTPUT_SCHEMA=schema.readable(SCHEMA[1]),
        AVAILABLE_PKGS=",".join(list(packages_distributions().keys()))
    )
    system = system.replace("\n"," ")

    messages = [
        {"role": "system", "content": system},
        {"role": "user", "content": prompt}
    ]

    if hist:
        hist.append({"role": "user", "content": prompt})
        messages = hist

        if compact_hist:
            # Saving tokens: keeping all system and user
            messages = [m for m in hist[:-2] if m["role"] in ["system","user"]]
            # and only the last assistant message, plus current message to send
            messages = messages + hist[-2:]

    response = aiclient.completions(provider, model, messages)

    hist = hist if hist else messages

    hist.append({
        "role": response.choices[0].message.role,
        "content": response.choices[0].message.content
    })

    return response.choices[0].message.content, hist


def update_code(code, file_name=FILE_NAME):
    with open(file_name, "w") as f:
        f.write(code)
    return None #hashlib.sha256(code.encode()).hexdigest()

def process_code(file_name=FILE_NAME, rerun=False):
    with open(file_name, "r") as f:
        code = f.read()

    hash_before = hashlib.sha256(remove_st_key_suffix(code).encode()).hexdigest()
    code = post_processing(code, rerun=rerun)
    hash_after = hashlib.sha256(remove_st_key_suffix(code).encode()).hexdigest()

    with open(file_name, "w") as f:
        f.write(code)

    return hash_before != hash_after

def generate_code(prompt, provider, model, system=SYSTEM, file_name=FILE_NAME, hist=None, compact_hist=False, rerun=False):
    new_code, hist = get_code(prompt, provider, model, system=SYSTEM, hist=hist, compact_hist=compact_hist)

    new_code = post_processing(new_code, rerun=rerun)

    digest = update_code(new_code, file_name=file_name)
    return hist, digest


def fix_code(error, provider, model, file_name=FILE_NAME, hist=None, compact_hist=False, rerun=False):
    with open(file_name) as f:
        code = f.read()
    code = preprocessing(code)
    prompt = (
        f"Fix: {error}."
        if hist and compact_hist else
        f"""The following piece of code:

```python
{code}
```

yields the following error:

```
{error}
```. Generate the new version of the code with no explanations and never include
markdown backticks. Only the code itself.
"""
    )

    hist, digest = generate_code(
        prompt,
        provider,
        model,
        system=SYSTEM,
        file_name=file_name,
        hist=hist,
        compact_hist=compact_hist,
        rerun=rerun
    )

    return hist, digest
    
def add_code(prompt, provider, model, file_name=FILE_NAME, hist=None, compact_hist=False, rerun=False):
    with open(file_name) as f:
        code = f.read()

    code = preprocessing(code)

    prompt = (
        f"Keep the code above and add the following features: {prompt}"
        if hist and compact_hist else
        f"""Keep the following piece of code:

```python
{code}
```

and add the following features:

```
{prompt}
```

Never modify the initial code, but simply merge it with the new features.
Generate the new version of the code with no explanations and never include
markdown backticks. Only the code itself. The new features must be appended as
the last tasks to be executed.

"""
        )
    hist, digest = generate_code(
        prompt,
        provider,
        model,
        system=SYSTEM,
        file_name=file_name,
        hist=hist,
        compact_hist=compact_hist,
        rerun=rerun
    )

    return hist, digest
    
def remove_code(prompt, provider, model, file_name=FILE_NAME, hist=None, compact_hist=False, rerun=False):
    with open(file_name) as f:
        code = f.read()
    
    code = preprocessing(code)

    prompt = (
        f"Remove the following features from the code above: {prompt}"
        if hist and compact_hist else
        f"""From the following piece of code:

```python
{code}
```

remove the following features:

```
{prompt}
```

Remove only what is referred to above. Never remove the other parts of the
code. Generate the new version of the code with no explanations and never
include markdown backticks. Only the code itself.

"""
    )
    hist, digest = generate_code(
        prompt,
        provider,
        model,
        system=SYSTEM,
        file_name=file_name,
        hist=hist,
        compact_hist=compact_hist,
        rerun=rerun
    )

    return hist, digest


def add_st_key_suffix(code, suffix="__k"):
    def gen_key_suffix(re_match):
        if suffix in re_match.group(2): # alternate pattern if suffix already there from past iteration
            new_id = re.sub(f'{suffix}[0-9]+',f'{suffix}{re_match.span()[1]}{re_match.span()[0]}',re_match.group(2))
            return f"{re_match.group(1)}{new_id}{re_match.group(3)}"

        return f"{re_match.group(1)}{re_match.group(2)}__k{re_match.span()[0]}{re_match.span()[1]}{re_match.group(3)}"

    return re.sub(r'(key=[\'"])([^"\']+)([\'"])', gen_key_suffix, code)

def remove_st_key_suffix(code, suffix="__k"):
    return re.sub(f"{suffix}[0-9]+", "", code)
    


def adapt_code_rerun(new_code, add=True):
    orig_form = f"def {FUNCTION_NAME}("
    proc_form = f"@st.fragment\ndef {FUNCTION_NAME}("

    if add and new_code.find(proc_form) < 0:
        new_code = new_code.replace(orig_form,proc_form)

    if not add:
        new_code = new_code.replace(proc_form,orig_form)

    return new_code

def post_processing(new_code, rerun):
    # removing annoying markdown backticks
    new_code = "\n".join(
        new_code.split("\n")[1:-1]
    ) if '```python' in new_code else new_code

    # making sure streamlit keys are unique
    new_code = add_st_key_suffix(new_code)

    # Adding code snippet for the rerun feature
    new_code = adapt_code_rerun(new_code, add=rerun)
        
    return new_code

def preprocessing(code):
    # removing rerun code if exists
    code = adapt_code_rerun(code, add=False)
    return code




import streamlit as st
import numpy as np
from PIL import Image

from autogui import autogui

st.set_page_config(layout="wide")

preview, pipeline = st.columns([0.7,0.3])

preview.subheader("Image")
tab_load, tab_proc, tab_orig = preview.tabs(["Upload", "Processed", "Original"])

def load_image() -> np.ndarray:
    """ {IO} """

    loaded_img = autogui("Image load testing tes", init_prompt="upload file")
    return loaded_img

with tab_load:
    img = load_image()


pipeline.subheader("Pipeline")


def proc_pipeline(img: np.ndarray) -> np.ndarray:
    """
    {IO}
    {VISUALIZATION}

Make sure to organize every step of the pipeline in an individual streamlit
expander. The image always must always go through every step of the pipeline
and return the final result.

Always add, for each expander, a component to toggle whether to apply that step
to the image. Never hide the components.

"""

    proc_img = autogui("Pipeline", history=autogui.STATIC, init_prompt="Brightness and contrast\n\nblack and white\n\ncrop")
    return proc_img

def view_img(img: np.ndarray) -> None:
    """ {IO} {VISUALIZATION} """

    autogui("img view", init_prompt="display image")
    return None




if isinstance(img,np.ndarray):
    proc = img.copy()

    with pipeline:
        proc = proc_pipeline(img=proc)

    proc = proc if isinstance(proc,type(img)) else img

    image = Image.fromarray(img)
    tab_orig.image(Image.fromarray(img), caption="Original", use_container_width=False)

    with tab_proc:
        view_img(proc)



from pathlib import Path

import setuptools

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setuptools.setup(
    name="streamlit-autogui",
    version="0.0.2",
    author="Caio Benatti Moretti",
    author_email="caiodba@gmail.com ",
    description="Vibe code inside a streamlit application and prompt for features on the fly.",
    long_description="Vibe code inside a streamlit application and prompt for technical implementations coupled with GUI components on the fly. Give tweaks on top of the generated code if needed.",
    long_description_content_type="text/plain",
    url="https://github.com/moretticb/streamlit-autogui",
    packages=setuptools.find_namespace_packages(),
#    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[],
    python_requires=">=3.7",
    install_requires=[
        "streamlit >= 0.63",
        "openai >= 2.6.1",
        "aisuite >= 0.1.13",
        "streamlit-code-editor >= 0.1.22"
    ],
    extras_require={
        "devel": [
            "wheel",
            "pytest==7.4.0",
            "playwright==1.48.0",
            "requests==2.31.0",
            "pytest-playwright-snapshot==1.0",
            "pytest-rerunfailures==12.0",
        ]
    }
)



import os
import streamlit.components.v1 as components
import streamlit as st
from code_editor import code_editor
import time
from . import aiclient
from . import aicodegen
from . import schema
from . import settings as s
from pathlib import Path
import re

import tempfile
import sys
import importlib.util


_RELEASE=True

if not _RELEASE:
    _component_func = components.declare_component(
        "autogui",
        url="http://localhost:3001"
    )
else:
    parent_dir = os.path.dirname(os.path.abspath(__file__))
    build_dir = os.path.join(parent_dir, "frontend/build")
    _component_func = components.declare_component("autogui", path=build_dir)

STATIC = 0 #1<<0
COMPACT = 1 #1<<2
FULL = 2 #1<<1

def autogui(
    name,
    init_prompt=None,
    provider=None,
    model=None,
    patience=3,
    rerun=True,
    history = COMPACT,
    like=None,
    args=None,
    system=None,
    system_fix=None,
    features=None,
    key=None,
    icon=":material/touch_app:"
):
    use_hist = history == STATIC
    compact_hist = history == COMPACT

    @st.dialog(f"AutoGUI {name}", width="medium")
    def gen_tool(filename):
        tabnames = [":material/code_blocks: Generate", ":material/build: Adjust", ":material/settings: Settings"]

        if filename.exists():
            placeholder = "Describe features to generate, add, or remove. You can prompt for fixes."
        else:
            tabnames = [t for t in tabnames if "Adjust" not in t]
            placeholder = "Describe features to generate..."

        tabs = st.tabs(tabnames)
        def tab(tabname, label=False):
            return next((tabnames[i] if label else tabs[i] for i,t in enumerate(tabnames) if tabname.lower() in t.lower()),None)

        with tab("generate"):
            try:
                current_model = f"Prompt ({": ".join(s.get_provider_model(key))})"
            except:
                current_model = f":material/error: Unable to find a valid model. Please provide a valid model in **{tab('settings', label=True)}**."
            prompt = st.text_area(current_model,placeholder=placeholder)

            btns = st.columns(4 if tab("adjust") else 1)

            cmd = {
                "generate" : btns[0].button("Generate", icon=":material/code_blocks:",key=f"{key}-gen", type="primary", use_container_width=True),
                "add" : False if not tab("adjust") else btns[1].button("Add", icon=":material/add:",key=f"{key}-add", type="secondary", use_container_width=True),
                "remove" : False if not tab("adjust") else btns[2].button("Remove", icon=":material/remove:",key=f"{key}-remove", type="secondary", use_container_width=True),
                "fix" : False if not tab("adjust") else btns[3].button("Fix", icon=":material/build:",key=f"{key}-fix", type="secondary", use_container_width=True)
            }

            for c in cmd:
                if cmd[c]:
                    provider,model = s.get_provider_model(key)

                    if c=='generate':
                        hist, digest = aicodegen.generate_code(prompt, provider, model, file_name=filename, compact_hist=compact_hist, rerun=not s.getpref("rerun",key))
                        st.session_state[hist_key] = hist if use_hist else None
                    else:
                        hist, digest = getattr(aicodegen,f"{c}_code")(prompt, provider, model, file_name=filename, hist=st.session_state[hist_key], compact_hist=compact_hist, rerun=not s.getpref("rerun",key))
                        st.session_state[hist_key] = hist

                    st.rerun()
                    break
            


        if tab("adjust"):
            with tab("adjust"):
                st.markdown("Generated code", help=f"at `{filename}`", unsafe_allow_html=True)
                with open(filename) as f:
                    code = f.read()
                    code = aicodegen.post_processing(code, not s.getpref("rerun", key))
                    resp = code_editor(code, lang="python", options={"wrap":True, "showLineNumbers":True})

                st.markdown(":material/build: <sub>Press \u2318+Enter to apply changes.</sub>", unsafe_allow_html=True)

                if resp.get("type") == "submit":
                    aicodegen.update_code(resp["text"], file_name=filename)
                    st.rerun()

        with tab("settings"):
            st.header("AI")
            c1,c2 = st.columns(2)
            try:
                def_provider, _ = s.get_provider_model(key)
            except:
                def_provider = None
            provider = c1.selectbox("Provider",[p for p in s.s.providers], index=0 if not def_provider else list(s.s.providers).index(def_provider))
            if provider:
                disabled = len(s.s.providers[provider]["warnings"])>0
                err = "\n\n:material/error: "
                err = err+err.join(s.s.providers[provider]["warnings"]) if disabled else ""
                st.markdown(err, unsafe_allow_html=True)
                deployment_name = c2.text_input("Model",value=s.s.providers[provider].get("default",""), disabled=disabled, help=s.s.providers[provider].get('help',''))


            c1,c2,c3 = st.columns(3)
            patience = c1.number_input("Patience", value=s.getpref("patience",key), min_value=1, max_value=100)

            hist_opt = ["Static","Compact","Full"]
            hist=c3.selectbox("Prompt history", hist_opt, index=s.getpref("history",key), help="Format of chat history. Choose `Static` for one-time or few requests, `Compact` for multiple chained requests (e.g., generate, then add/remove/fix multiple times), or `Full` to preserve the whole verbosy history. Notice that the longer the history, the more expensive it is (cost and time wise).")
            hist=hist_opt.index(hist)


            st.header("Streamlit")
            rerun = st.toggle("Refresh upon GUI change", value=s.getpref("rerun",key), help="If enabled, the whole dashboard updates after a GUI component has changed. Disable if many components must be adjusted before expecting an update. When disabled, a new button is added to effectiely apply changes when needed. Changing this setting might reset the dashboard state.")

            st.markdown("Apply changes to:")
            #st.header("Scope")
            #apply_to = s.GKEY if st.toggle("For all AutoGUI instances", value=True) else key
            apply_to = None
            
            c12,c3 = st.columns([0.3,0.7], vertical_alignment="center")
            c1,c2 = c12.columns([0.7,0.3], vertical_alignment="center", gap=None)
            name_label = s.getpref(key=key).instance
            name_label = name_label if len(name_label) < 13 else name_label[:13]+"..."
            if c1.button(name_label,key="autogui_apply_settings_local", use_container_width=True, disabled=disabled):
                apply_to=key
            if c2.button("All",key="autogui_apply_settings_global", use_container_width=True, disabled=disabled):
                apply_to=s.GKEY

            if apply_to:
                do_apply = True
                if not deployment_name:
                    do_apply=False
                    details = "" if "help" not in s.s.providers[provider] else f"Refer to {s.s.providers[provider]['help']} for more details."
                    c3.write(f"Model name not provided. {details}")

                if do_apply:
                    do_rerun = False

                    s.set_provider_model(provider, deployment_name, apply_to)
                    s.setpref("patience",patience,apply_to)
                    s.setpref("history",hist,apply_to)

                    do_rerun = s.getpref("rerun",apply_to) != rerun
                    s.setpref("rerun",rerun,apply_to)

                    c3.write(f"Settings applied to **{s.getpref(key=apply_to).instance}**.")
                    if do_rerun:
                        st.rerun()



    # applying initial global settings
    key, actual_init = s.init(name, key, provider, model, patience, rerun, history, st.session_state)
    hist_key = f"autogui-{key}-hist"
    settings_key = 'autogui-settings'


    digest=None
    #hashkey = f"autogui-{key}-digest"

    if like==None or args==None:
        name, invars, outvars, system, args = schema.from_parent_caller()
    else:
        name, invars, outvars = schema.from_func(like)

    #aicodegen.FUNCTION_NAME = name
    if system:
        aicodegen.SYSTEM = system
    if system_fix:
        aicodegen.SYSTEM_FIX = system_fix

    aicodegen.SCHEMA = (invars,outvars)

    file_key = f"autogui-{key}-code"
    if file_key not in st.session_state:
        st.session_state[file_key] = tempfile.TemporaryDirectory(delete=False)

    filename = Path(st.session_state[file_key].name) / aicodegen.FILE_NAME
    sys.path.insert(0, filename.parent)
    #module = str(filename.parent / filename.stem)
    module = str(filename.stem)

    display_icon = icon
    if not filename.exists() and init_prompt:
        try:
            provider,model = s.get_provider_model(key)
            hist, digest = aicodegen.generate_code(init_prompt, provider, model, file_name=filename, compact_hist=compact_hist)
            st.session_state[hist_key] = hist if use_hist else None
        except Exception as e:
            display_icon = ":material/error:"

    @st.fragment
    def open_dialog():
        if st.button("", icon=display_icon, key=f"{key}-autogui-btn", use_container_width=True):
            gen_tool(filename)

    open_dialog()
    _ = _component_func(name=name, key=key)

    component_value = None
    if len(outvars) > 1:
        component_value = (None for _ in outvars)


    gui_area = st.empty()
    error_area = st.empty()

    if filename.exists():
        p=0
        lacking_capabilities = []
        while p < patience:
            has_changed = aicodegen.process_code(file_name=filename, rerun=not s.getpref("rerun",key))
            if has_changed:
                st.rerun()
            try:
                spec = importlib.util.spec_from_file_location(aicodegen.FUNCTION_NAME,str(filename))
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)

                #st.session_state[hashkey] = digest
                with gui_area.container():
                    component_value = module.fcn(**args)
                    if not s.getpref("rerun",key) and st.button(f"Run {s.getpref(key=key).instance}",key=f"autogui-{key}-frag-apply"):
                        pass # Using button to rerun
#                    error_area.empty()
                break

            except Exception as e:
                if isinstance(e,ModuleNotFoundError):
                    lacking_capabilities.append(e.msg.split(' ')[-1])

                provider,model =  s.get_provider_model()
                aicodegen.fix_code(str(e), provider, model, file_name=filename, hist=st.session_state[hist_key], compact_hist=compact_hist)

                spec = importlib.util.spec_from_file_location(aicodegen.FUNCTION_NAME,str(filename))
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)

                p=p+1
#                st.write(p)
#                error_area.write(e)

        if len(lacking_capabilities) > 1:
            raise aicodegen.InsufficientCapabilityError(",".join(lacking_capabilities))
            

    #temp_dir.cleanup()
    return component_value

autogui.STATIC = STATIC
autogui.FULL = FULL
autogui.COMPACT = COMPACT
